@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2019 by Sukchan Lee <acetcom@gmail.com>
+ * Copyright (C) 2019-2024 by Sukchan Lee <acetcom@gmail.com>
  *
  * This file is part of Open5GS.
  *
@@ -111,7 +111,11 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
         enb = mme_enb_find_by_addr(addr);
         if (!enb) {
             enb = mme_enb_add(sock, addr);
-            ogs_assert(enb);
+            if (!enb) {
+                ogs_error("mme_enb_add() failed");
+                ogs_sock_destroy(sock);
+                ogs_free(addr);
+            }
         } else {
             ogs_warn("eNB context duplicated with IP-address [%s]!!!",
                     OGS_ADDR(addr, buf));
@@ -136,7 +140,10 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
         enb = mme_enb_find_by_addr(addr);
         if (!enb) {
             enb = mme_enb_add(sock, addr);
-            ogs_assert(enb);
+            if (!enb) {
+                ogs_error("amf_enb_add() failed");
+                ogs_free(addr);
+            }
         } else {
             ogs_free(addr);
         }
@@ -193,7 +200,7 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
 
         rc = ogs_s1ap_decode(&s1ap_message, pkbuf);
         if (rc == OGS_OK) {
-            e->enb = enb;
+            e->enb_id = enb->id;
             e->s1ap_message = &s1ap_message;
             ogs_fsm_dispatch(&enb->sm, e);
         } else {
@@ -210,13 +217,14 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
         break;
 
     case MME_EVENT_S1AP_TIMER:
-        enb_ue = e->enb_ue;
-        ogs_assert(enb_ue);
+        enb_ue = enb_ue_find_by_id(e->enb_ue_id);
+        if (!enb_ue) {
+            ogs_error("S1 Context has already been removed");
+            break;
+        }
 
         switch (e->timer_id) {
         case MME_TIMER_S1_DELAYED_SEND:
-            enb = e->enb;
-            ogs_assert(enb);
             pkbuf = e->pkbuf;
             ogs_assert(pkbuf);
 
@@ -239,23 +247,26 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
         break;
 
     case MME_EVENT_EMM_MESSAGE:
-        enb_ue = e->enb_ue;
-        ogs_assert(enb_ue);
         pkbuf = e->pkbuf;
         ogs_assert(pkbuf);
 
+        enb_ue = enb_ue_find_by_id(e->enb_ue_id);
+        if (!enb_ue) {
+            ogs_error("S1 Context has already been removed");
+            break;
+        }
+
         if (ogs_nas_emm_decode(&nas_message, pkbuf) != OGS_OK) {
             ogs_error("ogs_nas_emm_decode() failed");
             ogs_pkbuf_free(pkbuf);
             return;
         }
 
-        mme_ue = enb_ue->mme_ue;
+        mme_ue = mme_ue_find_by_id(enb_ue->mme_ue_id);
         if (!mme_ue) {
             mme_ue = mme_ue_find_by_message(&nas_message);
             if (!mme_ue) {
                 mme_ue = mme_ue_add(enb_ue);
-                MME_UE_CHECK(OGS_LOG_DEBUG, mme_ue);
                 if (mme_ue == NULL) {
                     r = s1ap_send_ue_context_release_command(enb_ue,
                             S1AP_Cause_PR_misc,
@@ -326,25 +337,23 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
                     enb_ue ? enb_ue->enb_ue_s1ap_id : 0,
                     enb_ue ? enb_ue->mme_ue_s1ap_id : 0);
             ogs_fatal("context [%p:%p]", enb_ue, mme_ue);
-            ogs_fatal("cycle [%p:%p]",
-                    enb_ue_cycle(enb_ue), mme_ue_cycle(mme_ue));
             ogs_fatal("IMSI [%s]", mme_ue ? mme_ue->imsi_bcd : "No MME_UE");
             ogs_assert_if_reached();
         }
         ogs_assert(OGS_FSM_STATE(&mme_ue->sm));
 
-        e->mme_ue = mme_ue;
+        e->mme_ue_id = mme_ue->id;
         e->nas_message = &nas_message;
 
         ogs_fsm_dispatch(&mme_ue->sm, e);
         if (OGS_FSM_CHECK(&mme_ue->sm, emm_state_exception)) {
-            mme_send_delete_session_or_mme_ue_context_release(mme_ue);
+            mme_send_delete_session_or_mme_ue_context_release(enb_ue, mme_ue);
         }
 
         ogs_pkbuf_free(pkbuf);
         break;
     case MME_EVENT_EMM_TIMER:
-        mme_ue = e->mme_ue;
+        mme_ue = mme_ue_find_by_id(e->mme_ue_id);
         ogs_assert(mme_ue);
         ogs_assert(OGS_FSM_STATE(&mme_ue->sm));
 
@@ -352,7 +361,7 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
         break;
 
     case MME_EVENT_ESM_MESSAGE:
-        mme_ue = e->mme_ue;
+        mme_ue = mme_ue_find_by_id(e->mme_ue_id);
         ogs_assert(mme_ue);
 
         pkbuf = e->pkbuf;
@@ -438,24 +447,20 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
      */
         if (OGS_FSM_CHECK(&mme_ue->sm, emm_state_de_registered)) {
             ESM_MESSAGE_CHECK;
-            MME_UE_CHECK(OGS_LOG_ERROR, mme_ue);
             ogs_pkbuf_free(pkbuf);
             break;
         } else if (OGS_FSM_CHECK(&mme_ue->sm, emm_state_authentication)) {
             ESM_MESSAGE_CHECK;
-            MME_UE_CHECK(OGS_LOG_ERROR, mme_ue);
             ogs_pkbuf_free(pkbuf);
             break;
         } else if (OGS_FSM_CHECK(&mme_ue->sm, emm_state_security_mode)) {
             ESM_MESSAGE_CHECK;
-            MME_UE_CHECK(OGS_LOG_ERROR, mme_ue);
             ogs_pkbuf_free(pkbuf);
             break;
         } else if (OGS_FSM_CHECK(&mme_ue->sm, emm_state_initial_context_setup)) {
         } else if (OGS_FSM_CHECK(&mme_ue->sm, emm_state_registered)) {
         } else if (OGS_FSM_CHECK(&mme_ue->sm, emm_state_exception)) {
             ESM_MESSAGE_CHECK;
-            MME_UE_CHECK(OGS_LOG_ERROR, mme_ue);
             ogs_pkbuf_free(pkbuf);
             break;
         }
@@ -467,12 +472,12 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
             break;
         }
 
-        sess = bearer->sess;
+        sess = mme_sess_find_by_id(bearer->sess_id);
         ogs_assert(sess);
         default_bearer = mme_default_bearer_in_sess(sess);
         ogs_assert(default_bearer);
 
-        e->bearer = bearer;
+        e->bearer_id = bearer->id;
         e->nas_message = &nas_message;
 
         ogs_fsm_dispatch(&bearer->sm, e);
@@ -508,7 +513,7 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
         break;
 
     case MME_EVENT_ESM_TIMER:
-        bearer = e->bearer;
+        bearer = mme_bearer_find_by_id(e->bearer_id);
         ogs_assert(bearer);
         ogs_assert(OGS_FSM_STATE(&bearer->sm));
 
@@ -516,6 +521,9 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
         break;
 
     case MME_EVENT_S6A_MESSAGE:
+        s6a_message = e->s6a_message;
+        ogs_assert(s6a_message);
+
         /*
          * A race condition can occur in the following situations.
          * In conclusion, we can use this situation to determine
@@ -557,14 +565,17 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
          * decide whether to process or
          * ignore the Authentication-Information-Answer as shown below.
          */
-        mme_ue = mme_ue_cycle(e->mme_ue);
+        mme_ue = mme_ue_find_by_id(e->mme_ue_id);
         if (!mme_ue) {
             ogs_error("UE(mme-ue) context has already been removed");
             goto cleanup;
         }
 
-        s6a_message = e->s6a_message;
-        ogs_assert(s6a_message);
+        enb_ue = enb_ue_find_by_id(e->enb_ue_id);
+        /*
+         * The 'enb_ue' context is not checked
+         * because the status is checked in the sending routine.
+         */
 
         switch (s6a_message->cmd_code) {
         case OGS_DIAM_S6A_CMD_CODE_AUTHENTICATION_INFORMATION:
@@ -573,12 +584,8 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
             if (emm_cause != OGS_NAS_EMM_CAUSE_REQUEST_ACCEPTED) {
                 ogs_info("[%s] Attach reject [OGS_NAS_EMM_CAUSE:%d]",
                         mme_ue->imsi_bcd, emm_cause);
-                enb_ue = enb_ue_cycle(mme_ue->enb_ue);
-                if (!enb_ue) {
-                    ogs_error("S1 context has already been removed");
-                    break;
-                }
-                r = nas_eps_send_attach_reject(mme_ue, emm_cause,
+                r = nas_eps_send_attach_reject(
+                        enb_ue, mme_ue, emm_cause,
                         OGS_NAS_ESM_CAUSE_PROTOCOL_ERROR_UNSPECIFIED);
                 ogs_expect(r == OGS_OK);
                 ogs_assert(r != OGS_ERROR);
@@ -594,26 +601,32 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
             ogs_debug("OGS_DIAM_S6A_CMD_CODE_UPDATE_LOCATION");
             emm_cause = mme_s6a_handle_ula(mme_ue, s6a_message);
             if (emm_cause != OGS_NAS_EMM_CAUSE_REQUEST_ACCEPTED) {
-                ogs_info("[%s] Attach reject [OGS_NAS_EMM_CAUSE:%d]",
-                        mme_ue->imsi_bcd, emm_cause);
-                enb_ue = enb_ue_cycle(mme_ue->enb_ue);
-                if (!enb_ue) {
-                    ogs_error("S1 context has already been removed");
-                    break;
-                }
-                r = nas_eps_send_attach_reject(mme_ue, emm_cause,
-                        OGS_NAS_ESM_CAUSE_PROTOCOL_ERROR_UNSPECIFIED);
-                ogs_expect(r == OGS_OK);
-                ogs_assert(r != OGS_ERROR);
+                if (mme_ue->nas_eps.type == MME_EPS_TYPE_ATTACH_REQUEST) {
+                    ogs_info("[%s] Attach reject [OGS_NAS_EMM_CAUSE:%d]",
+                            mme_ue->imsi_bcd, emm_cause);
+                    r = nas_eps_send_attach_reject(
+                            enb_ue, mme_ue, emm_cause,
+                            OGS_NAS_ESM_CAUSE_PROTOCOL_ERROR_UNSPECIFIED);
+                    ogs_expect(r == OGS_OK);
+                    ogs_assert(r != OGS_ERROR);
+                } else if (mme_ue->nas_eps.type == MME_EPS_TYPE_TAU_REQUEST) {
+                    ogs_info("[%s] TAU reject [OGS_NAS_EMM_CAUSE:%d]",
+                            mme_ue->imsi_bcd, emm_cause);
+                    r = nas_eps_send_tau_reject(
+                            enb_ue, mme_ue, emm_cause);
+                    ogs_expect(r == OGS_OK);
+                    ogs_assert(r != OGS_ERROR);
+                } else
+                    ogs_error("Invalid Type[%d]", mme_ue->nas_eps.type);
 
                 r = s1ap_send_ue_context_release_command(enb_ue,
                         S1AP_Cause_PR_nas, S1AP_CauseNas_normal_release,
-                        S1AP_UE_CTX_REL_UE_CONTEXT_REMOVE, 0);
+                        mme_ue_find_by_id(enb_ue->mme_ue_id) ?
+                            S1AP_UE_CTX_REL_UE_CONTEXT_REMOVE :
+                            S1AP_UE_CTX_REL_S1_CONTEXT_REMOVE, 0);
                 ogs_expect(r == OGS_OK);
                 ogs_assert(r != OGS_ERROR);
             }
-
-            mme_ue->location_updated_but_not_canceled_yet = true;
             break;
         case OGS_DIAM_S6A_CMD_CODE_PURGE_UE:
             ogs_debug("OGS_DIAM_S6A_CMD_CODE_PURGE_UE");
@@ -621,7 +634,6 @@ void mme_state_operational(ogs_fsm_t *s, mme_event_t *e)
             break;
         case OGS_DIAM_S6A_CMD_CODE_CANCEL_LOCATION:
             ogs_debug("OGS_DIAM_S6A_CMD_CODE_CANCEL_LOCATION");
-            mme_ue->location_updated_but_not_canceled_yet = false;
             mme_s6a_handle_clr(mme_ue, s6a_message);
             break;
         case OGS_DIAM_S6A_CMD_CODE_INSERT_SUBSCRIBER_DATA:
@@ -768,10 +780,16 @@ cleanup:
         break;
 
     case MME_EVENT_S11_TIMER:
-        sgw_ue = e->sgw_ue;
-        ogs_assert(sgw_ue);
-        mme_ue = sgw_ue->mme_ue;
-        ogs_assert(mme_ue);
+        sgw_ue = sgw_ue_find_by_id(e->sgw_ue_id);
+        if (!sgw_ue) {
+            ogs_error("SGW-UE Context has already been removed");
+            break;
+        }
+        mme_ue = mme_ue_find_by_id(sgw_ue->mme_ue_id);
+        if (!mme_ue) {
+            ogs_error("MME-UE Context has already been removed");
+            break;
+        }
 
         switch (e->timer_id) {
         case MME_TIMER_S11_HOLDING:
@@ -784,10 +802,15 @@ cleanup:
                 GTP_COUNTER_INCREMENT(
                     mme_ue, GTP_COUNTER_DELETE_SESSION_BY_PATH_SWITCH);
 
-                ogs_assert(OGS_OK ==
-                    mme_gtp_send_delete_session_request(
-                        sgw_ue, sess,
-                        OGS_GTP_DELETE_IN_PATH_SWITCH_REQUEST));
+                enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
+                if (enb_ue) {
+                    ogs_assert(OGS_OK ==
+                        mme_gtp_send_delete_session_request(
+                            enb_ue, sgw_ue, sess,
+                            OGS_GTP_DELETE_IN_PATH_SWITCH_REQUEST));
+                } else
+                    ogs_error("ENB-S1 Context has already been removed");
+
             }
             break;
 
@@ -836,6 +859,11 @@ cleanup:
             mme_gn_handle_sgsn_context_request(xact, &gtp1_message.sgsn_context_request);
             break;
         case OGS_GTP1_SGSN_CONTEXT_RESPONSE_TYPE:
+            /* Clang scan-build SA: NULL pointer dereference: mme_ue=NULL if both gtp1_message.h.teid=0 and
+             * xact->local_teid=0. The following function mme_gn_handle_sgsn_context_response() handles the NULL
+             * but the later calls to OGS_FSM_TRAN() to change state will be a NULL pointer dereference. */
+            ogs_assert(mme_ue);
+
             /* 3GPP TS 23.401 Figure D.3.6-1 step 5 */
             rv = mme_gn_handle_sgsn_context_response(xact, mme_ue, &gtp1_message.sgsn_context_response);
             if (rv == OGS_GTP1_CAUSE_ACCEPT) {
@@ -860,9 +888,9 @@ cleanup:
         break;
 
     case MME_EVENT_GN_TIMER:
-        mme_ue = e->mme_ue;
+        mme_ue = mme_ue_find_by_id(e->mme_ue_id);
         ogs_assert(mme_ue);
-        sgw_ue = mme_ue->sgw_ue;
+        sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
         ogs_assert(sgw_ue);
 
         switch (e->timer_id) {
@@ -877,10 +905,15 @@ cleanup:
             ogs_list_for_each(&mme_ue->sess_list, sess) {
                 GTP_COUNTER_INCREMENT(
                     mme_ue, GTP_COUNTER_DELETE_SESSION_BY_PATH_SWITCH);
-                ogs_assert(OGS_OK ==
-                    mme_gtp_send_delete_session_request(
-                        sgw_ue, sess,
-                        OGS_GTP_DELETE_IN_PATH_SWITCH_REQUEST));
+
+                enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
+                if (enb_ue) {
+                    ogs_assert(OGS_OK ==
+                        mme_gtp_send_delete_session_request(
+                            enb_ue, sgw_ue, sess,
+                            OGS_GTP_DELETE_IN_PATH_SWITCH_REQUEST));
+                } else
+                    ogs_error("ENB-S1 Context has already been removed");
             }
             break;
 
